﻿Option Explicit On
Option Strict On

#Region "参照設定"

Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Imports System.Drawing
Imports System.Windows.Forms

Imports Box2DLib
Imports Box2DLib.Settings
Imports Box2DLib.MathFunction

#End Region

'//////////////////////////////////////////////////////////////
''' <summary>
''' Box2D用のフォームクラスです。
''' </summary>
''' <remarks></remarks>
Public Class Box2DForm
    Inherits D3DForm

    '==========================================================

#Region "初期化・破棄関連"

    ''' <summary>
    ''' クラスを初期化します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()

    End Sub

    ''' <summary>
    ''' 内部オブジェクトを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overrides Sub Release()

        'Box2D関連の開放
        DestroyWorld()

        'DirectX関連の開放
        MyBase.Release()

    End Sub

#End Region

    '==========================================================

#Region "シミュレーションワールド関係"

    '内部データ

    ''' <summary>シミュレーションワールドを保持します。</summary>
    Private mWorld As bb2World

    ''' <summary>シミュレーション実行時のタイムステップを保持します。</summary>
    Protected mTimeStep As Single = 1.0F / 60.0F

    ''' <summary>シミュレーション実行時の繰り返し回数を保持します。</summary>
    Protected mIterations As Integer = 10

    'プロパティ

    ''' <summary>
    ''' シミュレーションワールドを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns>現在のシュミレーションワールド</returns>
    ''' <remarks>セットアップされてない場合は、Nothingが返されます。</remarks>
    Public ReadOnly Property World() As bb2World
        Get
            Return mWorld
        End Get
    End Property

    ''' <summary>
    ''' シミュレーション実行時のタイムステップを取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property TimeStep() As Single
        Get
            Return mTimeStep
        End Get
        Set(ByVal value As Single)
            mTimeStep = value
        End Set
    End Property

    ''' <summary>
    ''' シミュレーション実行時の繰り返し回数を取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Iterations() As Integer
        Get
            Return mIterations
        End Get
        Set(ByVal value As Integer)
            mIterations = value
        End Set
    End Property

    'メソッド

    ''' <summary>
    ''' シミュレーションワールドを生成します。
    ''' </summary>
    ''' <param name="WorldSizeX">シミュレーションワールドの大きさ（横）</param>
    ''' <param name="WorldSizeY">シミュレーションワールドの大きさ（縦）</param>
    ''' <param name="Gravity">重力係数</param>
    ''' <returns>成功した場合False、失敗した場合True</returns>
    ''' <remarks></remarks>
    Public Overridable Function CreateWorld(ByVal WorldSizeX As Single, ByVal WorldSizeY As Single, ByVal Gravity As Single) As Boolean

        '現在のワールドを掃き
        DestroyWorld()

        'ワールド生成
        Dim waabb As New bb2AABB
        waabb.lowerBound.Set(-WorldSizeX, -WorldSizeY)
        waabb.upperBound.Set(WorldSizeX, WorldSizeY)
        Dim grav As New bb2Vec2(0, Gravity)
        mWorld = New bb2World(waabb, grav, True)

    End Function

    ''' <summary>
    ''' シミュレーションワールドを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub DestroyWorld()

        'ボディリストのクリア
        For Each bdy As BodyWrapper In mBodyList
            bdy.Release()
        Next
        mBodyList.Clear()

        'ジョイントリストのクリア
        For Each jnt As JointWrapper In mJointList
            jnt.Release()
        Next
        mJointList.Clear()

        'シミュレーションワールドの破棄
        If mWorld IsNot Nothing Then
            mWorld.Dispose()
            mWorld = Nothing
        End If

    End Sub

    ''' <summary>
    ''' シミュレーション時間を進めます。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub WorldStep()

        'チェック
        If mWorld Is Nothing Then Exit Sub

        'シミュレーション実施
        mWorld.Step(mTimeStep, mIterations)

    End Sub

#End Region

#Region "ボディリスト/ジョイントリスト関連"

    '内部データ

    ''' <summary>表示対象ボディリストを保持します。</summary>
    Protected mBodyList As New List(Of BodyWrapper)

    ''' <summary>描画対象のジョイントリストを保持します。</summary>
    Protected mJointList As New List(Of JointWrapper)

    'プロパティ

    ''' <summary>
    ''' 表示対象ボディのリストを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property BodyList() As List(Of BodyWrapper)
        Get
            Return mBodyList
        End Get
    End Property

    ''' <summary>
    ''' 表示対象ジョイントのリストを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property JointList() As List(Of JointWrapper)
        Get
            Return mJointList
        End Get
    End Property

    'メソッド

    ''' <summary>
    ''' 新しいボディを生成し、ボディリストに登録します。
    ''' </summary>
    ''' <param name="PosX">位置ｘ</param>
    ''' <param name="PosY">位置ｙ</param>
    ''' <param name="Angle">角度(単位：ラジアン）</param>
    ''' <returns>生成されワールドに登録されたボディをラップするボディラッパ</returns>
    ''' <remarks></remarks>
    Public Function CreateBody(Optional ByVal PosX As Single = 0, Optional ByVal PosY As Single = 0, Optional ByVal Angle As Single = 0) As BodyWrapper

        'ボディ及びボディラッパ生成
        Dim bdy As BodyWrapper = New BodyWrapper(mWorld, PosX, PosY, Angle)

        'リストに追加
        mBodyList.Add(bdy)

        Return bdy
    End Function

    ''' <summary>
    ''' ジョイントをジョイントリストに登録します。
    ''' </summary>
    ''' <param name="Joint">Box2Dジョイント</param>
    ''' <param name="LineColor">線色</param>
    ''' <returns>ジョイントラッパ</returns>
    ''' <remarks>
    ''' 線色がTransparentの場合、ジョイントリストへの登録は行われずNothingが返されます。
    ''' </remarks>
    Public Function CreateJoint(ByVal Joint As bb2Joint, ByVal LineColor As Color) As JointWrapper

        '引数チェック
        If Joint Is Nothing Then Return Nothing
        If LineColor = Color.Transparent Then Return Nothing

        'ジョイントラッパ作成
        Dim jw As New JointWrapper(Joint, LineColor)

        'ジョイントリストへの登録
        mJointList.Add(jw)

        Return jw
    End Function

    ''' <summary>
    ''' 回転ジョイントを生成します。
    ''' </summary>
    ''' <param name="Body1"></param>
    ''' <param name="Body2"></param>
    ''' <param name="AnchorX"></param>
    ''' <param name="AnchorY"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function CreateRevoluteJoint(ByVal Body1 As BodyWrapper, ByVal Body2 As BodyWrapper, ByVal AnchorX As Single, ByVal AnchorY As Single, ByVal LineColor As Color, Optional ByVal collideConnected As Boolean = True) As JointWrapper

        'ジョイント定義の作成
        Dim jd As New bb2RevoluteJointDef
        jd.Initialize(Body1.Body, Body2.Body, New bb2Vec2(AnchorX, AnchorY))
        jd.collideConnected = collideConnected

        'ジョイントの登録
        Dim jnt As bb2Joint = mWorld.CreateJoint(jd)

        'ジョイントリストに登録
        Return CreateJoint(jnt, LineColor)
    End Function

    ''' <summary>
    ''' ボディリを検索します。
    ''' </summary>
    ''' <param name="Body">探し出すボディ</param>
    ''' <returns>見付かった場合対象のボディラッパ。見付からなかった場合Nothing</returns>
    ''' <remarks></remarks>
    Public Function FindBody(ByVal Body As bb2Body) As BodyWrapper

        For Each bw As BodyWrapper In mBodyList
            If bw.Body.BaseObject = Body.BaseObject Then
                '発見
                Return bw
            End If
        Next

        '見付からず
        Return Nothing
    End Function

#End Region

#Region "基本動作"

    'イベント定義

    Public Event OnStep()

    '内部データ

    ''' <summary>モーダル表示終了時の戻り値を保持します。</summary>
    Protected mReturnCode As Integer

    ''' <summary>一時停止中かどうかの状態を保持します。</summary>
    Protected mPause As Boolean

    'プロパティ

    ''' <summary>
    ''' 一時停止中かどうかを取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Pause() As Boolean
        Get
            Return mPause
        End Get
        Set(ByVal value As Boolean)
            mPause = value
        End Set
    End Property

    'メソッド

    ''' <summary>
    ''' フォームを表示し、ユーザ操作を受け付けます。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub DoModal()

        'セットアップチェック
        If mDev Is Nothing Then Exit Sub 'セットアップが完了していない

        '表示
        Me.Show()
        Me.Refresh()

        'メインループ
        Do While Me.Created

            'シミュレーション
            If Not mPause Then

                'イベント発生
                RaiseEvent OnStep()

                'シミュレーション進行
                WorldStep()

            End If

            '描画処理
            Render()

            'メッセージ処理
            Application.DoEvents()

        Loop

    End Sub

    ''' <summary>
    ''' 描画時の処理です。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub OnDraw()

        '描画
        DrawAll()

        'イベント発生
        MyBase.OnDraw()

    End Sub

    ''' <summary>
    ''' ボディリストに登録されているボディを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub DrawAll()

        'ボディの描画
        For Each bdy As BodyWrapper In mBodyList
            bdy.Draw()
        Next

        'ジョイントの描画
        ResetWorldTransform()
        For Each jnt As JointWrapper In mJointList
            jnt.Draw()
        Next

    End Sub

#End Region

    '==========================================================

#Region "ユーザ操作"

    '定数
    Public Const ZOOMMIN As Single = 0.25F
    Public Const DEFAULTZOOMSPEED As Single = 1.1F

    '内部データ

    ''' <summary>画面のズーム係数を保持します。</summary>
    Protected mZoom As Single = 1.0F

    ''' <summary>ズーム速度を保持します。</summary>
    Protected mZoomSpeed As Single = DEFAULTZOOMSPEED

    ''' <summary>中心オフセットを保持します。</summary>
    Protected mCenterOffset As PointF = New PointF(0, 0)

    ''' <summary>マウス右ボタン押下時のマウス位置を保持します。</summary>
    Protected mMouseDownPoint As Point

    ''' <summary>マウス右ボタン押下時の表示中心位置を保持します。</summary>
    Protected mMouseDownCenter As PointF

    ''' <summary>操作用のマウスジョイントを保持します。</summary>
    Private mMouseJoint As bb2MouseJoint

    ''' <summary>操作用マウスジョイントのラッパを保持します。</summary>
    Private mMouseJointWrapper As JointWrapper

    'プロパティ

    ''' <summary>
    ''' 画面表示のズーム係数を取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks>例外が発生する可能性があります。(0を指定すると例外が発生します)</remarks>
    Public Property Zoom() As Single
        Get
            Return mZoom
        End Get
        Set(ByVal value As Single)

            'チェックと補正
            If value < ZOOMMIN Then value = ZOOMMIN

            '更新
            mZoom = value

            'サイズ変更処理
            SetupView()

        End Set
    End Property

    ''' <summary>
    ''' ズーム速度を取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property ZoomSpeed() As Single
        Get
            Return mZoomSpeed
        End Get
        Set(ByVal value As Single)
            mZoomSpeed = value
        End Set
    End Property

    ''' <summary>
    ''' 中心オフセットを取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property CenterOffset() As PointF
        Get
            Return mCenterOffset
        End Get
        Set(ByVal value As PointF)

            '値の保持
            mCenterOffset = value

            'サイズ変更処理
            SetupView()

        End Set
    End Property

    'メソッド

    ''' <summary>
    ''' 表示サイズ変更時の処理を行います。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overrides Sub SetupView()

        '状態チェック
        If mDev Is Nothing Then Exit Sub

        Dim sz As Size = Me.ClientSize
        Dim mat As Matrix = Matrix.Identity

        '縦横比を合わせる
        If sz.Width > sz.Height Then
            mat.M11 = CSng(sz.Height / sz.Width)
            mat.M22 = 1
        Else
            mat.M11 = 1
            mat.M22 = CSng(sz.Width / sz.Height)
        End If

        '平衡移動
        mat.M41 = mCenterOffset.X * mat.M11
        mat.M42 = mCenterOffset.Y * mat.M22

        'ズーム
        mat.Multiply(Matrix.Scaling(1 / mZoom, 1 / mZoom, 1))

        '射影変換行列の設定
        mDev.Transform.Projection = mat

    End Sub

    ''' <summary>
    ''' マウスジョイントを作成します。
    ''' </summary>
    ''' <param name="MousePosition">マウスの位置</param>
    ''' <remarks></remarks>
    Private Sub CreateMouseJoint(ByVal MousePosition As Point)

        If mMouseJoint IsNot Nothing Then
            'マウスジョイントの削除
            DeleteMouseJoint()
        End If

        'マウス座標をシミュレーションワールド座標に変換
        Dim mp As Vector2 = ConvertMouseToWorld(MousePosition)
        Dim pp As New bb2Vec2(mp.X, mp.Y)

        '検索範囲の決定
        Dim dp As New bb2Vec2(0.001F, 0.001F)
        Dim aabb As New bb2AABB
        aabb.lowerBound = pp - dp
        aabb.upperBound = pp + dp

        'マウス位置付近に存在するシェイプを検索
        Const maxShapes As Integer = 32
        Dim shp(maxShapes - 1) As bb2Shape
        Dim cnt As Integer = mWorld.Query(aabb, shp, maxShapes)

        'マウス位置にあるシェイプの特定
        Dim target As bb2Body = Nothing
        For t As Integer = 0 To cnt - 1
            Dim bdy As bb2Body = shp(t).GetBody()
            If (bdy.IsStatic() = False) And (bdy.GetMass() > 0.0F) Then
                If shp(t).TestPoint(bdy.GetXForm(), pp) Then
                    target = bdy
                    Exit For
                End If
            End If
        Next
        If target Is Nothing Then Exit Sub '対象なし

        'マウスジョイントの作成
        Dim md As New bb2MouseJointDef
        md.body1 = mWorld.GetGroundBody()
        md.body2 = target
        md.target = pp
        md.maxForce = 1000.0F * target.GetMass()
        Dim jt As bb2Joint = mWorld.CreateJoint(md)
        mMouseJoint = bb2MouseJoint.CastFrom(jt)

        'ジョイントリストへの登録
        mMouseJointWrapper = New JointWrapper(jt, Color.LightGray)
        mJointList.Add(mMouseJointWrapper)

    End Sub

    ''' <summary>
    ''' マウスジョイント移動させます。
    ''' </summary>
    ''' <param name="MousePosition">マウスの位置</param>
    ''' <remarks></remarks>
    Private Sub MoveMouseJoint(ByVal MousePosition As Point)

        'チェック
        If mMouseJoint Is Nothing Then Exit Sub

        'マウス座標をシミュレーションワールド座標に変換
        Dim mp As Vector2 = ConvertMouseToWorld(MousePosition)
        Dim wp As New bb2Vec2(mp.X, mp.Y)

        '位置更新
        mMouseJoint.SetTarget(wp)

    End Sub

    ''' <summary>
    ''' マウスジョイントを削除します。
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub DeleteMouseJoint()

        If mMouseJoint IsNot Nothing Then

            'ジョイントリストからの削除
            mJointList.Remove(mMouseJointWrapper)

            'ジョイントの削除
            mWorld.DestroyJoint(mMouseJoint)
            mMouseJoint = Nothing

        End If

    End Sub

    'イベントハンドラ

    ''' <summary>
    ''' マウスボタンが押された時の処理です。
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Box2DForm_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown

        If (e.Button And Windows.Forms.MouseButtons.Left) <> Windows.Forms.MouseButtons.None Then
            '左ボタン押下時

            'マウスジョイントを作成
            CreateMouseJoint(e.Location)

        End If

        If (e.Button And Windows.Forms.MouseButtons.Right) <> Windows.Forms.MouseButtons.None Then
            '右ボタン押下時

            '現在の位置を保存
            mMouseDownPoint = e.Location
            mMouseDownCenter = mCenterOffset

        End If

    End Sub

    ''' <summary>
    ''' マウスが移動された時の処理です。
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Box2DForm_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove

        If (e.Button And Windows.Forms.MouseButtons.Left) <> Windows.Forms.MouseButtons.None Then
            '左ボタン押下時

            'マウスジョイントの移動
            MoveMouseJoint(e.Location)

        End If

        If (e.Button And Windows.Forms.MouseButtons.Right) <> Windows.Forms.MouseButtons.None Then
            '右ボタン押下時

            'マウス座標をワールド座標に変更
            Dim p1 As Vector2 = ConvertMouseToWorld(mMouseDownPoint)
            Dim p2 As Vector2 = ConvertMouseToWorld(e.Location)

            '表示中心座標を更新
            Me.CenterOffset = New PointF(mMouseDownCenter.X + p2.X - p1.X, mMouseDownCenter.Y + p2.Y - p1.Y)

        End If

    End Sub

    ''' <summary>
    ''' マウスボタンが離された時の処理です。
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Box2DForm_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp

        If (e.Button And Windows.Forms.MouseButtons.Left) <> Windows.Forms.MouseButtons.None Then
            '左ボタン押下時

            'マウスジョイントの削除
            DeleteMouseJoint()

        End If

    End Sub

    ''' <summary>
    ''' マウスホイール回転時の処理です。
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Box2DForm_MouseWheel(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseWheel

        If mZoomSpeed <> 0 Then
            If e.Delta < 0 Then
                Me.Zoom *= mZoomSpeed
            Else
                Me.Zoom /= mZoomSpeed
            End If
        End If

    End Sub

    ''' <summary>
    ''' キー入力時の処理です。
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Box2DForm_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown

        Select Case e.KeyCode
            Case Keys.P, Keys.Pause
                mPause = Not mPause

        End Select

    End Sub

#End Region

    '==========================================================

#Region "座標変換関連"

    ''' <summary>
    ''' スクリーン座標をワールド座標に変換します。
    ''' </summary>
    ''' <param name="PosX">スクリーン座標ｘ</param>
    ''' <param name="PosY">スクリーン座標ｙ</param>
    ''' <returns>ワールド座標</returns>
    ''' <remarks></remarks>
    Public Function ConvertScreenToWorld(ByVal PosX As Single, ByVal PosY As Single) As Vector2
        Dim ret As New Vector2(PosX, PosY)
        ret.TransformNormal(Matrix.Invert(mDev.Transform.Projection))
        ret.X -= mCenterOffset.X
        ret.Y -= mCenterOffset.Y
        Return ret
    End Function

    ''' <summary>
    ''' スクリーン座標をワールド座標に変換します。
    ''' </summary>
    ''' <param name="ScreenPos">スクリーン座標</param>
    ''' <returns>ワールド座標</returns>
    ''' <remarks></remarks>
    Public Function ConvertScreenToWorld(ByVal ScreenPos As PointF) As Vector2
        Return ConvertScreenToWorld(ScreenPos.X, ScreenPos.Y)
    End Function

    ''' <summary>
    ''' マウス座標をスクリーン座標に変換します。
    ''' </summary>
    ''' <param name="PosX">マウス座標ｘ</param>
    ''' <param name="PosY">マウス座標ｙ</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function ConvertMouseToScreen(ByVal PosX As Integer, ByVal PosY As Integer) As PointF
        Dim sx As Single = CSng(Me.ClientSize.Width / 2)
        Dim sy As Single = CSng(Me.ClientSize.Height / 2)
        Return New PointF((PosX - sx) / sx, (sy - PosY) / sy)
    End Function

    ''' <summary>
    ''' マウス座標をスクリーン座標に変換します。
    ''' </summary>
    ''' <param name="MousePos">マウス座標</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function ConvertMouseToScreen(ByVal MousePos As Point) As PointF
        Return ConvertMouseToScreen(MousePos.X, MousePos.Y)
    End Function

    ''' <summary>
    ''' マウス座標をワールド座標に変換します。
    ''' </summary>
    ''' <param name="MousePos">マウス座標</param>
    ''' <returns>ワールド座標</returns>
    ''' <remarks></remarks>
    Public Function ConvertMouseToWorld(ByVal MousePos As Point) As Vector2
        Dim sp As PointF = ConvertMouseToScreen(MousePos)
        Return ConvertScreenToWorld(sp)
    End Function

    ''' <summary>
    ''' マウス座標をワールド座標に変換します。
    ''' </summary>
    ''' <param name="PosX">マウス座標ｘ</param>
    ''' <param name="PosY">マウス座標ｙ</param>
    ''' <returns>ワールド座標</returns>
    ''' <remarks></remarks>
    Public Function ConvertMouseToWorld(ByVal PosX As Integer, ByVal PosY As Integer) As Vector2
        Return ConvertScreenToWorld(ConvertMouseToScreen(PosX, PosY))
    End Function

#End Region

End Class
