﻿Option Explicit On
Option Strict On

#Region "参照設定関連"

Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Imports System.Drawing
Imports System.Windows.Forms

#End Region

'//////////////////////////////////////////////////////////////
''' <summary>
''' Direct3Dで2Dプリミティブを描画する為の基底クラスです。
''' </summary>
''' <remarks></remarks>
Public MustInherit Class D3DPrimitive2D

    '==========================================================

#Region "Direct3Dデバイス関連"

    ''' <summary>Direct3Dデバイスを保持します。</summary>
    Protected Shared mDev As Microsoft.DirectX.Direct3D.Device

    ''' <summary>
    ''' Direct3Dデバイスを取得・設定します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Property Device() As Microsoft.DirectX.Direct3D.Device
        Get
            Return mDev
        End Get
        Set(ByVal value As Microsoft.DirectX.Direct3D.Device)
            mDev = value
        End Set
    End Property

#End Region

#Region "生成・開放関連"

    ''' <summary>
    ''' プリミティブを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub Release()

        If mVB IsNot Nothing Then
            mVB.Dispose()
            mVB = Nothing
        End If

    End Sub

#End Region

#Region "頂点バッファ関連"

    ''' <summary>頂点バッファを保持します。</summary>
    Protected mVB As Microsoft.DirectX.Direct3D.VertexBuffer

    ''' <summary>
    ''' 頂点バッファを生成します。
    ''' </summary>
    ''' <param name="VertexCount">頂点数</param>
    ''' <remarks></remarks>
    Protected MustOverride Sub CreateVertexBuffer(ByVal VertexCount As Integer)

    ''' <summary>
    ''' 頂点バッファを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overridable Sub DeleteVertexBuffer()

        If mVB IsNot Nothing Then
            mVB.Dispose()
            mVB = Nothing
        End If

    End Sub

    ''' <summary>
    ''' 頂点バッファをロックし、書き込み用のGraphicStreamを返します。
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Protected Function Lock() As GraphicsStream
        Return mVB.Lock(0, 0, LockFlags.None)
    End Function

    ''' <summary>
    ''' 頂点バッファのロックを解除します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Sub Unlock()
        mVB.Unlock()
    End Sub

#End Region

#Region "描画関連"

    ''' <summary>
    ''' プリミティブを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public MustOverride Sub Draw()

#End Region

End Class


'//////////////////////////////////////////////////////////////
''' <summary>
''' Direct3Dで単色2Dプリミティブを描画する為のクラスです。
''' </summary>
''' <remarks></remarks>
Public Class D3DPrimitive2DSolid
    Inherits D3DPrimitive2D

    '==========================================================

#Region "頂点バッファ関連"

    '内部データ

    ''' <summary>頂点のリストを保持します。</summary>
    Protected mVertexList As New List(Of CustomVertex.PositionColored)

    ''' <summary>描画時のプリミティブ数を保持します。</summary>
    Protected mPrimitiveCount As Integer

    ''' <summary>描画時の輪郭線数を保持します。</summary>
    Protected mBorderLineCount As Integer

    'メソッド

    ''' <summary>
    ''' 頂点バッファを生成します。
    ''' </summary>
    ''' <param name="VertexCount">頂点数</param>
    ''' <remarks></remarks>
    Protected Overrides Sub CreateVertexBuffer(ByVal VertexCount As Integer)

        mVB = New VertexBuffer(GetType(CustomVertex.PositionColored), VertexCount, mDev, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed)

    End Sub

    ''' <summary>
    ''' 頂点バッファを生成します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overridable Overloads Sub CreateVertexBuffer()

        '頂点バッファの生成
        CreateVertexBuffer(mVertexList.Count)

        '頂点情報の転送
        UpdateVertex()

    End Sub

    ''' <summary>
    ''' 頂点バッファの内容を更新します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overridable Sub UpdateVertex()

        Dim gs As GraphicsStream = Lock()
        For Each v As CustomVertex.PositionColored In mVertexList
            gs.Write(v)
        Next

    End Sub

#End Region

#Region "描画関連"

    ''' <summary>
    ''' プリミティブを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overrides Sub Draw()

        'チェック
        If mVB Is Nothing Then Throw New Exception("頂点バッファが生成されていません。")
        If mDev Is Nothing Then Throw New Exception("Direct3Dデバイスが準備されていません。")

        '頂点バッファの選択
        mDev.SetStreamSource(0, mVB, 0)
        mDev.VertexFormat = CustomVertex.PositionColored.Format

        'ポリゴン描画
        Dim bidx As Integer = 0
        If mPrimitiveCount > 0 Then
            mDev.DrawPrimitives(PrimitiveType.TriangleFan, 0, mPrimitiveCount)
            bidx = mPrimitiveCount + 2
        End If

        '境界線描画
        If mBorderLineCount > 0 Then
            mDev.DrawPrimitives(PrimitiveType.LineStrip, bidx, mBorderLineCount)
        End If

    End Sub

#End Region

#Region "作成関連"

    ''' <summary>
    ''' 四角形のプリミティブを作成します。
    ''' </summary>
    ''' <param name="SizeX">横幅</param>
    ''' <param name="SizeY">縦幅</param>
    ''' <param name="FillColor">塗り潰し色</param>
    ''' <param name="BorderColor">輪郭線色</param>
    ''' <param name="OffsetX">オフセットｘ</param>
    ''' <param name="OffsetY">オフセットｙ</param>
    ''' <param name="PosZ">ｚ座標</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CreateBox(ByVal SizeX As Single, ByVal SizeY As Single, ByVal FillColor As Color, ByVal BorderColor As Color, Optional ByVal OffsetX As Single = 0, Optional ByVal OffsetY As Single = 0, Optional ByVal PosZ As Single = 0) As D3DPrimitive2DSolid

        'オブジェクト生成
        Dim obj As New D3DPrimitive2DSolid

        '頂点作成準備
        Dim hx As Single = SizeX / 2.0F
        Dim hy As Single = SizeY / 2.0F
        Dim c1 As Integer = FillColor.ToArgb
        Dim c2 As Integer = BorderColor.ToArgb

        '塗り潰し部分の頂点データ作成
        If FillColor <> Color.Transparent Then
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX - hx, OffsetX + hy, PosZ, c1))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + hx, OffsetX + hy, PosZ, c1))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + hx, OffsetX - hy, PosZ, c1))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX - hx, OffsetX - hy, PosZ, c1))
            obj.mPrimitiveCount = 2
        End If

        '境界線用の頂点データ作成
        If BorderColor <> Color.Transparent Then
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX - hx, OffsetX + hy, PosZ, c2))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + hx, OffsetX + hy, PosZ, c2))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + hx, OffsetX - hy, PosZ, c2))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX - hx, OffsetX - hy, PosZ, c2))
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX - hx, OffsetX + hy, PosZ, c2))
            obj.mBorderLineCount = 4
        End If

        '頂点バッファ作成
        obj.CreateVertexBuffer()

        '頂点情報更新
        obj.UpdateVertex()

        Return obj
    End Function

    ''' <summary>
    ''' 多角形のプリミティブを生成します。
    ''' </summary>
    ''' <param name="Pnt">頂点のリスト</param>
    ''' <param name="FillColor">塗り潰し色</param>
    ''' <param name="BorderColor">輪郭線色</param>
    ''' <param name="OffsetX">オフセットｘ</param>
    ''' <param name="OffsetY">オフセットｙ</param>
    ''' <param name="PosZ">ｚ座標</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CreatePolygon(ByVal Pnt As List(Of PointF), ByVal FillColor As Color, ByVal BorderColor As Color, Optional ByVal OffsetX As Single = 0, Optional ByVal OffsetY As Single = 0, Optional ByVal PosZ As Single = 0) As D3DPrimitive2DSolid

        '引数チェック
        If Pnt Is Nothing Then Return Nothing
        If Pnt.Count < 3 Then Return Nothing

        'オブジェクト生成
        Dim obj As New D3DPrimitive2DSolid

        '頂点作成準備
        Dim c1 As Integer = FillColor.ToArgb
        Dim c2 As Integer = BorderColor.ToArgb

        '塗り潰し部分の頂点データ作成
        If FillColor <> Color.Transparent Then
            For Each p As PointF In Pnt
                obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + p.X, OffsetX + p.Y, PosZ, c1))
            Next
            obj.mPrimitiveCount = Pnt.Count - 2
        End If

        '境界線用の頂点データ作成
        If BorderColor <> Color.Transparent Then
            For Each p As PointF In Pnt
                obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + p.X, OffsetX + p.Y, PosZ, c2))
            Next
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX + Pnt(0).X, OffsetX + Pnt(0).Y, PosZ, c2))
            obj.mBorderLineCount = Pnt.Count
        End If

        '頂点バッファ作成
        obj.CreateVertexBuffer()

        '頂点情報更新
        obj.UpdateVertex()

        Return obj
    End Function

    ''' <summary>
    ''' 線プリミティブを生成します。
    ''' </summary>
    ''' <param name="LineCount">線の数</param>
    ''' <param name="LineColor">線色</param>
    ''' <param name="PosZ">ｚ座標</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CreateLine(ByVal LineCount As Integer, ByVal LineColor As Color, Optional ByVal OffsetX As Single = 0, Optional ByVal OffsetY As Single = 0, Optional ByVal PosZ As Single = 0) As D3DPrimitive2DSolid

        'オブジェクト生成
        Dim obj As New D3DPrimitive2DSolid

        '頂点作成準備
        Dim c1 As Integer = LineColor.ToArgb

        '塗り潰し部分の頂点データ作成
        obj.mPrimitiveCount = 0

        '線用の頂点データ作成
        For t As Integer = 0 To LineCount
            obj.mVertexList.Add(New CustomVertex.PositionColored(0, 0, PosZ, c1))
        Next
        obj.mBorderLineCount = LineCount

        '頂点バッファ作成
        obj.CreateVertexBuffer()

        '頂点情報更新
        obj.UpdateVertex()

        Return obj
    End Function

    ''' <summary>
    ''' 円形(多角形近似)のプリミティブを生成します。
    ''' </summary>
    ''' <param name="Radius">半径</param>
    ''' <param name="FillColor">塗り潰し色</param>
    ''' <param name="BorderColor">輪郭線色</param>
    ''' <param name="OffsetX">オフセットｘ</param>
    ''' <param name="OffsetY">オフセットｙ</param>
    ''' <param name="PosZ">ｚ座標</param>
    ''' <param name="VertexCount">頂点数</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CreateCircle(ByVal Radius As Single, ByVal FillColor As Color, ByVal BorderColor As Color, Optional ByVal OffsetX As Single = 0, Optional ByVal OffsetY As Single = 0, Optional ByVal PosZ As Single = 0, Optional ByVal VertexCount As Integer = 32) As D3DPrimitive2DSolid

        'オブジェクト生成
        Dim obj As New D3DPrimitive2DSolid

        '頂点作成準備
        Dim c1 As Integer = FillColor.ToArgb
        Dim c2 As Integer = BorderColor.ToArgb
        If VertexCount <= 2 Then
            VertexCount = 32
        End If

        '頂点座標計算
        Dim pnt(VertexCount) As PointF
        For t As Integer = 0 To VertexCount
            Dim r As Single = CSng(t * 2 * 3.141592 / VertexCount)
            If t = VertexCount Then r += 0.001F 'ポリゴン欠けを防ぐ為の処置
            pnt(t).X = OffsetX + Radius * CSng(Math.Cos(r))
            pnt(t).Y = OffsetY + Radius * CSng(Math.Sin(r))
        Next

        '塗り潰し部分の頂点データ作成
        If FillColor <> Color.Transparent Then
            obj.mVertexList.Add(New CustomVertex.PositionColored(OffsetX, OffsetX, PosZ, c1))
            For t As Integer = 0 To VertexCount
                obj.mVertexList.Add(New CustomVertex.PositionColored(pnt(t).X, pnt(t).Y, PosZ, c1))
            Next
            obj.mPrimitiveCount = VertexCount
        End If

        '境界線用の頂点データ作成
        If BorderColor <> Color.Transparent Then
            For t As Integer = 0 To VertexCount
                obj.mVertexList.Add(New CustomVertex.PositionColored(pnt(t).X, pnt(t).Y, PosZ, c2))
            Next
            obj.mPrimitiveCount = VertexCount
        End If

        '頂点バッファ作成
        obj.CreateVertexBuffer()

        '頂点情報更新
        obj.UpdateVertex()

        Return obj
    End Function

#End Region

End Class

'//////////////////////////////////////////////////////////////
''' <summary>
''' Direct3Dで単色2D線プリミティブを描画する為のクラスです。
''' </summary>
''' <remarks></remarks>
Public Class D3DPrimitive2DLine
    Inherits D3DPrimitive2D

    ''' <summary>
    ''' 線情報を格納する構造体です。
    ''' </summary>
    ''' <remarks></remarks>
    Public Class LineData

        ''' <summary>始点座標</summary>
        Public Point1 As New CustomVertex.PositionColored

        ''' <summary>終点座標</summary>
        Public Point2 As New CustomVertex.PositionColored

        'メソッド

        ''' <summary>
        ''' 頂点１情報を更新します。
        ''' </summary>
        ''' <param name="PosX">座標ｘ</param>
        ''' <param name="PosY">座標ｙ</param>
        ''' <remarks></remarks>
        Public Sub SetPoint1(ByVal PosX As Single, ByVal PosY As Single)
            Point1.X = PosX
            Point1.Y = PosY
        End Sub

        ''' <summary>
        ''' 頂点２情報を更新します。
        ''' </summary>
        ''' <param name="PosX">座標ｘ</param>
        ''' <param name="PosY">座標ｙ</param>
        ''' <remarks></remarks>
        Public Sub SetPoint2(ByVal PosX As Single, ByVal PosY As Single)
            Point2.X = PosX
            Point2.Y = PosY
        End Sub

        ''' <summary>
        ''' グラフィックストリームに頂点データを書き出します。
        ''' </summary>
        ''' <param name="gs">グラフィックストリーム</param>
        ''' <remarks></remarks>
        Public Sub Write(ByVal gs As GraphicsStream)
            gs.Write(Point1)
            gs.Write(Point2)
        End Sub

    End Class

    '==========================================================

#Region "初期化"

    '内部データ

    ''' <summary>線リストを保持します。</summary>
    Protected mLineList As New List(Of LineData)

    ''' <summary>描画時のプリミティブ数を保持します。</summary>
    Protected mPrimitiveCount As Integer

    'プロパティ

    ''' <summary>
    ''' 線リストを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks>
    ''' 変更した線情報を有効にするには、Updateを呼び出す必要があります。
    ''' 頂点を削除したり追加することは禁止です。
    ''' </remarks>
    Public ReadOnly Property LineList() As List(Of LineData)
        Get
            Return mLineList
        End Get
    End Property

    'メソッド

    ''' <summary>
    ''' クラスを初期化します。
    ''' </summary>
    ''' <param name="LineCount">線の数</param>
    ''' <param name="LineColor">線の色</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal LineCount As Integer, ByVal LineColor As Color)

        '線リストの作成
        mLineList.Clear()
        mPrimitiveCount = LineCount
        Dim cc As Integer = LineColor.ToArgb
        For t As Integer = 0 To LineCount - 1
            Dim ld As New LineData
            ld.Point1.Color = cc
            ld.Point2.Color = cc
            mLineList.Add(ld)
        Next

        '頂点バッファ作成
        CreateVertexBuffer(mPrimitiveCount * 2)

    End Sub

    ''' <summary>
    ''' 頂点バッファを生成します。
    ''' </summary>
    ''' <param name="VertexCount">頂点数</param>
    ''' <remarks></remarks>
    Protected Overrides Sub CreateVertexBuffer(ByVal VertexCount As Integer)

        mVB = New VertexBuffer(GetType(CustomVertex.PositionColored), VertexCount, mDev, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed)

    End Sub

    ''' <summary>
    ''' 線情報を更新します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Update()

        '頂点バッファのロック
        Dim gs As GraphicsStream = Me.Lock()

        '頂点情報の書き出し

        For Each ld As LineData In mLineList
            ld.Write(gs)
        Next

        '頂点バッファの解除
        Me.Unlock()

    End Sub

#End Region

#Region "描画関連"

    ''' <summary>
    ''' プリミティブを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Overrides Sub Draw()

        'チェック
        If mVB Is Nothing Then Exit Sub
        If mLineList.Count <> mPrimitiveCount Then Throw New Exception("線の数が変更されました。")

        '頂点バッファの選択
        mDev.SetStreamSource(0, mVB, 0)
        mDev.VertexFormat = CustomVertex.PositionColored.Format

        '境界線描画
        If mPrimitiveCount > 0 Then
            mDev.DrawPrimitives(PrimitiveType.LineList, 0, mPrimitiveCount)
        End If

    End Sub

#End Region

End Class
