﻿Option Explicit On
Option Strict On

#Region "参照設定"

Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Imports System.Drawing
Imports System.Windows.Forms

Imports Box2DLib
Imports Box2DLib.Settings
Imports Box2DLib.MathFunction

#End Region

'//////////////////////////////////////////////////////////////
''' <summary>
''' Box2DのBodyオブジェクトをラップするクラスです。
''' </summary>
''' <remarks></remarks>
Public Class BodyWrapper

    '==========================================================

#Region "初期化・破棄関連"

    '内部データ

    ''' <summary>対象のBox2Dボディを保持します。</summary>
    Protected mBody As bb2Body

    ''' <summary>描画スタイルを保持します。</summary>
    Protected mDrawStyle As New DrawStyleData

    ''' <summary>ボディを構成するシェイプのリストを保持します。</summary>
    Protected mShapeList As New List(Of ShapeWrapper)

    'プロパティ

    ''' <summary>
    ''' ラップしているBox2Dボディオブジェクトを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Body() As bb2Body
        Get
            Return mBody
        End Get
    End Property

    ''' <summary>
    ''' 描画スタイルを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property DrawStyle() As DrawStyleData
        Get
            Return mDrawStyle
        End Get
    End Property

    'メソッド

    ''' <summary>
    ''' 作成済みのボディをラップします。
    ''' </summary>
    ''' <param name="Body">対象ボディ</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal Body As bb2Body, Optional ByVal AutoCreateShapeList As Boolean = True)

        '引数チェック
        If Body Is Nothing Then Throw New Exception("対象のBodyが指定されていません。")

        '引数保存
        mBody = Body

        'シェイプ情報の取得
        If AutoCreateShapeList Then
            CreateShapeList()
        End If

    End Sub

    ''' <summary>
    ''' 新しいボディを作成します。
    ''' </summary>
    ''' <param name="World">シミュレーションワールド</param>
    ''' <param name="PosX">位置ｘ</param>
    ''' <param name="PosY">位置ｙ</param>
    ''' <param name="Angle">角度(ラジアン)</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal World As bb2World, Optional ByVal PosX As Single = 0, Optional ByVal PosY As Single = 0, Optional ByVal Angle As Single = 0)

        '引数チェック
        If World Is Nothing Then Throw New Exception("対象のWorldが指定されていません。")

        'ボディ定義の作成
        Dim bdydef As New bb2BodyDef
        bdydef.position.Set(PosX, PosY)
        bdydef.angle = Angle

        'ボディ生成
        mBody = World.CreateBody(bdydef)

    End Sub

    ''' <summary>
    ''' ボディを構成するシェイプを列挙し、シェイプリストを作成します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Sub CreateShapeList()

        '現在のシェイプリストを破棄
        ReleaseShapeList()

        '先頭のシェイプ取得
        Dim shp As bb2Shape = mBody.GetShapeList()

        '全てのシェイプを列挙
        Do While shp IsNot Nothing

            'シェイプラッパ作成
            Dim sp As New ShapeWrapper(shp, mDrawStyle)

            'リストに追加
            mShapeList.Add(sp)

            '次のシェイプへ
            shp = shp.GetNext()
        Loop

    End Sub

    ''' <summary>
    ''' 内部のデータを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Release()

        'シェイプリストの破棄
        ReleaseShapeList()

        'ボディの開放
        If mBody IsNot Nothing Then
            mBody.Dispose()
        End If
        mBody = Nothing

    End Sub

    ''' <summary>
    ''' シェイプリストを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Sub ReleaseShapeList()

        For Each shp As ShapeWrapper In mShapeList
            shp.Release()
        Next
        mShapeList.Clear()

    End Sub

    ''' <summary>
    ''' 表示用シェイプ情報を更新します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub UpdateShape()

        'シェイプリスト作成
        CreateShapeList()

    End Sub

#End Region

#Region "描画関連"

    ''' <summary>
    ''' ボディを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Draw()

        '移動と回転
        Dim pos As bb2Vec2 = mBody.GetPosition()
        Dim mt As Matrix = Matrix.Translation(pos.x, pos.y, 0)
        Dim mr As Matrix = Matrix.RotationZ(mBody.GetAngle())
        mr.Multiply(mt)
        D3DPrimitive2D.Device.Transform.World = mr

        '各シェイプの描画
        For Each shp As ShapeWrapper In mShapeList
            shp.Draw()
        Next

    End Sub

#End Region

#Region "生成関係"

    ''' <summary>
    ''' シェイプを追加します。
    ''' </summary>
    ''' <param name="ShapeDef">シェイプ定義</param>
    ''' <remarks></remarks>
    Public Function AddShape(ByVal ShapeDef As bb2ShapeDef) As ShapeWrapper

        'シェイプ登録
        Dim shp As bb2Shape = mBody.CreateShape(ShapeDef)

        '塊情報更新
        mBody.SetMassFromShapes()

        'シェイプラッパ作成
        Dim shpwp As ShapeWrapper = New ShapeWrapper(shp, mDrawStyle)

        'リストに追加
        mShapeList.Add(shpwp)

        Return shpwp
    End Function

    ''' <summary>
    ''' 矩形シェイプを追加します。
    ''' </summary>
    ''' <param name="SizeX">横幅</param>
    ''' <param name="SizeY">縦幅</param>
    ''' <param name="OffsetX">オフセット横</param>
    ''' <param name="OffsetY">オフセット縦</param>
    ''' <param name="Angle">角度(ラジアン)</param>
    ''' <param name="Density">密度</param>
    ''' <param name="Friction">摩擦係数</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function AddBox(ByVal SizeX As Single, ByVal SizeY As Single, Optional ByVal OffsetX As Single = 0, Optional ByVal OffsetY As Single = 0, Optional ByVal Angle As Single = 0, Optional ByVal Density As Single = 0, Optional ByVal Friction As Single = 0.2F, Optional ByVal Restitution As Single = 0.0F) As ShapeWrapper

        'シェイプ定義作成
        Dim off As New bb2Vec2(OffsetX, OffsetY)
        Dim shpdef As New bb2PolygonDef
        shpdef.density = Density
        shpdef.friction = Friction
        shpdef.restitution = Restitution
        shpdef.SetAsBox(SizeX, SizeY, off, Angle)

        'シェイプ登録
        Return AddShape(shpdef)
    End Function

    ''' <summary>
    ''' 円形シェイプを追加します。
    ''' </summary>
    ''' <param name="Radius">半径</param>
    ''' <param name="Density">密度</param>
    ''' <param name="Friction">摩擦係数</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function AddCircle(ByVal Radius As Single, Optional ByVal Density As Single = 0, Optional ByVal Friction As Single = 0.2F, Optional ByVal Restitution As Single = 0.0F) As ShapeWrapper

        'シェイプ定義作成
        Dim shpdef As New bb2CircleDef
        shpdef.density = Density
        shpdef.friction = Friction
        shpdef.restitution = Restitution
        shpdef.radius = Radius

        'シェイプ登録
        Return AddShape(shpdef)
    End Function

#End Region

End Class


'//////////////////////////////////////////////////////////////
''' <summary>
''' 描画時のスタイル(色等)を管理するクラスです。
''' </summary>
''' <remarks></remarks>
Public Class DrawStyleData

    ''' <summary>塗り潰し色</summary>
    Public FillColor As Color = Color.White

    ''' <summary>線色(輪郭線色)</summary>
    Public LineColor As Color = Color.Black

    ''' <summary>ｚ座標</summary>
    Public PosZ As Single = 0.0F

End Class
