﻿Option Explicit On
Option Strict On

#Region "参照設定関連"

Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Imports System.Drawing
Imports System.Windows.Forms

Imports Box2DLib
Imports Box2DLib.Settings
Imports Box2DLib.MathFunction

#End Region


'//////////////////////////////////////////////////////////////
''' <summary>
''' Box2DのJointオブジェクトをラップするクラスです。
''' </summary>
''' <remarks></remarks>
Public Class JointWrapper

    '==========================================================

#Region "初期化・破棄関連"

    '内部データ

    ''' <summary>ラップするBox2Dジョイントを保持します。</summary>
    Protected mJoint As bb2Joint

    'プロパティ

    ''' <summary>
    ''' アタッチされているBox2Dジョイントを返します。
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Joint() As bb2Joint
        Get
            Return mJoint
        End Get
    End Property

    'メソッド

    ''' <summary>
    ''' クラスを初期化します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New(ByVal Joint As bb2Joint, ByVal LineColor As Color)

        '引数チェック
        If Joint Is Nothing Then Throw New Exception("対象のJointが指定されていません。")

        '引数の保持
        mJoint = Joint
        mDrawStyle = New DrawStyleData
        mDrawStyle.LineColor = LineColor

        '描画用のプリミティブ準備
        CreatePrimitive()

    End Sub

    ''' <summary>
    ''' 内部データを破棄します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Release()

        If mPrimitive IsNot Nothing Then
            mPrimitive.Release()
            mPrimitive = Nothing
        End If
        If mJoint IsNot Nothing Then
            mJoint.Dispose()
        End If
        mJoint = Nothing

    End Sub

    ''' <summary>
    ''' 描画用のプリミティブを準備します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overridable Sub CreatePrimitive()

        'プリミティブ作成
        Select Case mJoint.GetType()
            Case bb2JointType.e_distanceJoint
                mPrimitive = New D3DPrimitive2DLine(1, mDrawStyle.LineColor)  '線は１本

            Case Else
                mPrimitive = New D3DPrimitive2DLine(2, mDrawStyle.LineColor)  '線は２本

        End Select

    End Sub

#End Region

#Region "描画関連"

    '内部データ

    ''' <summary>描画スタイルを保持します。</summary>
    Protected mDrawStyle As DrawStyleData

    ''' <summary>描画用プリミティブを保持します。</summary>
    Protected mPrimitive As D3DPrimitive2DLine

    'メソッド

    ''' <summary>
    ''' シェイプを描画します。
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Draw()

        '状態チェック
        If mPrimitive Is Nothing Then Exit Sub

        '座標更新
        Update()

        '描画
        mPrimitive.Draw()

    End Sub

    ''' <summary>
    ''' 頂点情報を更新します。
    ''' </summary>
    ''' <remarks></remarks>
    Protected Sub Update()

        '座標の更新
        Dim b1 As bb2Body = Joint.GetBody1()
        Dim b2 As bb2Body = Joint.GetBody2()
        Dim r1 As bb2Vec2 = b1.GetPosition()
        Dim r2 As bb2Vec2 = b2.GetPosition()
        Dim p1 As bb2Vec2 = Joint.GetAnchor1()
        Dim p2 As bb2Vec2 = Joint.GetAnchor2()
        Select Case mJoint.GetType()
            Case bb2JointType.e_distanceJoint
                mPrimitive.LineList(0).SetPoint1(p1.x, p1.y)
                mPrimitive.LineList(0).SetPoint2(p2.x, p2.y)

            Case bb2JointType.e_pulleyJoint
                Dim pj As bb2PulleyJoint = bb2PulleyJoint.CastFrom(mJoint)
                r1 = pj.GetGroundAnchor1()
                r2 = pj.GetGroundAnchor2()
                mPrimitive.LineList(0).SetPoint1(r1.x, r1.y)
                mPrimitive.LineList(0).SetPoint2(p1.x, p1.y)
                mPrimitive.LineList(1).SetPoint1(r2.x, r2.y)
                mPrimitive.LineList(1).SetPoint2(p2.x, p2.y)

            Case bb2JointType.e_mouseJoint
                mPrimitive.LineList(0).SetPoint1(p1.x, p1.y)
                mPrimitive.LineList(0).SetPoint2(p2.x, p2.y)

            Case Else
                mPrimitive.LineList(0).SetPoint1(r1.x, r1.y)
                mPrimitive.LineList(0).SetPoint2(p1.x, p1.y)
                mPrimitive.LineList(1).SetPoint1(r2.x, r2.y)
                mPrimitive.LineList(1).SetPoint2(p2.x, p2.y)

        End Select

        '頂点バッファの更新
        mPrimitive.Update()

    End Sub

#End Region

End Class
